#include "MQTTTelemetry.h"

#include <ArduinoJson.h>

#include "../../Static.h"
#include "../../Config.h"
#include "../../Platform.h"

static StaticJsonDocument<1024> m_json;

struct MQTTDevice {
  const String id;
  const String name;
  const String model;
  const String softwareVersion;
  const String manufacturer;
  const String availabilityTopic;

  void toJson(const JsonObject& json) const {
    json["name"] = name;
    json["mdl"] = model;
    json["sw"] = softwareVersion;
    json["mf"] = manufacturer;
    json["ids"][0] = id;
  }
};

const String availTopic = String("renderbug/") + Platform::deviceID() + "/availability";

const MQTTDevice Device{
  Platform::deviceID(),
  Platform::deviceName(),
  Platform::model(),
#ifdef ESP8266
  ESP.getSketchMD5(),
#else
  "",
#endif
  "Phong Robotics",
  availTopic
};

struct MQTTEntity {
  const MQTTDevice& device;
  String name;
  String entityId;
  String rootTopic;

  MQTTEntity(const String& domain, const MQTTDevice& device, const String& name) : device(device), name(Platform::deviceName() + " " + name) {
    entityId = String(device.id) + "-" + name;
    rootTopic = String("homeassistant/") + domain + String("/renderbug/") + entityId;
  }

  String configTopic() const {
    return rootTopic + "/config";
  }

  String commandTopic() const {
    return rootTopic + "/set";
  }

  String heartbeatTopic() const {
    return String("renderbug/") + Device.id + "/heartbeat";
  }

  String statTopic() const {
    return rootTopic + "/state";
  }

  bool isCommandTopic(const char* topic) const {
    if (strncmp(topic, rootTopic.c_str(), rootTopic.length()) == 0) {
      return strncmp(&topic[rootTopic.length()], "/set", sizeof("/set")) == 0;
    }
    return false;
  }

  void toJson(JsonDocument& jsonBuf, bool isInteractive = true) const {
    jsonBuf["~"] = rootTopic.c_str();
    jsonBuf["name"] = name;
    jsonBuf["unique_id"] = entityId;
    if (isInteractive) {
      jsonBuf["cmd_t"] = "~/set";
      jsonBuf["ret"] = true;
      jsonBuf["schema"] = "json";
    } else {
    }
    jsonBuf["stat_t"] = "~/state";
    jsonBuf["json_attr_t"] = heartbeatTopic();
    jsonBuf["avty_t"] = device.availabilityTopic;
    device.toJson(jsonBuf.createNestedObject("dev"));
  }
};

const MQTTEntity Lightswitch {
  "light", Device, "lightswitch"
};

const MQTTEntity FPSSensor {
  "sensor", Device, "fps"
};

MQTTTelemetry::MQTTTelemetry() : BufferedInputSource("MQTT"),
    m_mqtt(m_wifi),
    m_logPrinter(this)
{
    m_debugTopic = String("renderbug/") + Platform::deviceID();
    memset(m_hostBuf, 0, sizeof(m_hostBuf));
}

void
MQTTTelemetry::publishScenes()
{
  for(auto scene : Static<Sequencer>::instance()->scenes()) {
    m_json.clear();
    String strName{scene.name};
    MQTTEntity sceneObj {
      "scene",
      Device,
      strName
    };
    sceneObj.toJson(m_json, false);
    m_json["cmd_t"] = "~/set";
    m_json["ret"] = false;
    m_json["payload_on"] = "active";
    publishDoc(sceneObj.configTopic().c_str(), false);
    m_mqtt.subscribe(sceneObj.commandTopic().c_str());
    Log.info("Published scene %s", sceneObj.configTopic().c_str());
  }
}

void
MQTTTelemetry::onMQTTOnline()
{
  Log.notice("Connected to MQTT");
  String logTopic = m_debugTopic + "/log";
  Log.info("MQTT logs are available at %s", logTopic.c_str());
  m_needHeartbeat = true;

  m_json.clear();
  Lightswitch.toJson(m_json);
  m_json["brightness"] = true;
  m_json["rgb"] = true;

  publishDoc(Lightswitch.configTopic().c_str(), true);
  m_mqtt.subscribe(Lightswitch.commandTopic().c_str());

  publishScenes();

  m_json.clear();
  FPSSensor.toJson(m_json, false);
  m_json["unit_of_meas"] = "Frames/s";
  publishDoc(FPSSensor.configTopic().c_str(), true);

#ifdef ESP8266
  struct rst_info resetInfo = *ESP.getResetInfoPtr();
  if (resetInfo.reason != 0) {
    char buff[200];
    sprintf(&buff[0], "Fatal exception:%d flag:%d (%s) epc1:0x%08x epc2:0x%08x epc3:0x%08x excvaddr:0x%08x depc:0x%08x", resetInfo.exccause, resetInfo.reason, (resetInfo.reason == 0 ? "DEFAULT" : resetInfo.reason == 1 ? "WDT" : resetInfo.reason == 2 ? "EXCEPTION" : resetInfo.reason == 3 ? "SOFT_WDT" : resetInfo.reason == 4 ? "SOFT_RESTART" : resetInfo.reason == 5 ? "DEEP_SLEEP_AWAKE" : resetInfo.reason == 6 ? "EXT_SYS_RST" : "???"), resetInfo.epc1, resetInfo.epc2, resetInfo.epc3, resetInfo.excvaddr, resetInfo.depc); 
    Log.warning("Previous crash detected! %s", buff);
  }
#endif
}

void
MQTTTelemetry::handleEventOnline(const InputEvent& evt)
{
  if (!m_mqtt.connected()) {
    Log.notice("Connecting to MQTT as %s on %s...", Platform::deviceID(), Device.availabilityTopic.c_str());
    if (m_mqtt.connect(Platform::deviceID(), NULL, NULL, Device.availabilityTopic.c_str(), 0, true, "offline")) {
      onMQTTOnline();
    } else {
      Log.warning("Could not connect to MQTT");
    }
  } else {
    String statTopic = Lightswitch.statTopic();
    if (evt.intent == InputEvent::SetPower) {
      m_json.clear();
      m_isOn = evt.asInt() ? true : false;
      m_json["state"] = m_isOn ? "ON" : "OFF";
      publishDoc(statTopic.c_str());
    } else if (evt.intent == InputEvent::SetBrightness) {
      m_json.clear();
      m_json["brightness"] = evt.asInt();
      m_json["state"] = m_isOn ? "ON" : "OFF";
      publishDoc(statTopic.c_str());
    } else if (evt.intent == InputEvent::SetColor) {
      CRGB color = evt.asRGB();
      m_json.clear();
      m_json["color"]["r"] = color.r;
      m_json["color"]["g"] = color.g;
      m_json["color"]["b"] = color.b;
      m_json["state"] = m_isOn ? "ON" : "OFF";
      publishDoc(statTopic.c_str());
    } else if (evt.intent == InputEvent::SetScene) {
      m_json.clear();
      m_json["state"] = m_isOn ? "ON" : "OFF";
      publishDoc(statTopic.c_str());
    }
  }
}

void
MQTTTelemetry::handleConfigChange(const Configuration& cfg)
{
  Log.notice("Config change in mqtt");
  const char* newIP = cfg.get("mqtt.ip", m_hostBuf);
  Log.notice("ip: %s", newIP);
  strncpy(m_hostBuf, newIP, sizeof(m_hostBuf));
  m_hostBuf[sizeof(m_hostBuf)-1] = 0;
  m_mqtt.disconnect();
}

void
MQTTTelemetry::loop()
{
  BufferedInputSource::loop();
  OnlineTaskMixin::loop();
  ConfigTaskMixin::loop();
}

void
MQTTTelemetry::handleEvent(const InputEvent& evt)
{
  BufferedInputSource::handleEvent(evt);
  OnlineTaskMixin::handleEvent(evt);
  ConfigTaskMixin::handleEvent(evt);
}

void
MQTTTelemetry::onOnline()
{
  m_needHeartbeat = true;
  if (strlen(m_hostBuf) > 0) {
    m_mqtt.setServer(m_hostBuf, 1883);
  } else {
    Log.warning("mqtt.ip is not configured, no connection possible");
  }
  m_mqtt.setBufferSize(1024);
  m_mqtt.setCallback(&MQTTTelemetry::s_callback);
}

void
MQTTTelemetry::onOffline()
{
  m_mqtt.disconnect();
}

void
MQTTTelemetry::loopOnline()
{
  m_mqtt.loop();
  EVERY_N_SECONDS(10) {
    m_needHeartbeat = true;
  }
  if (m_needHeartbeat) {
    publishHeartbeat();
    m_needHeartbeat = false;
  }
}

void
MQTTTelemetry::publishHeartbeat()
{
  m_json.clear();
  m_json["device_id"] = Platform::deviceID();
  m_json["sketch_version"] = ESP.getSketchMD5();
  m_json["os_version"] = Platform::version();
  m_json["localip"] = WiFi.localIP().toString();
  m_json["pixelCount"] = Static<ConfigService>::instance()->coordMap()->physicalPixelCount();
  m_json["loadedProfile"] = Static<ConfigService>::instance()->loadedProfile();
  m_json["RSSI"] = WiFi.RSSI();
  m_json["free_ram"] = Platform::freeRam();
  m_json["fps"] = FastLED.getFPS();
  publishDoc(Lightswitch.heartbeatTopic().c_str());
  m_mqtt.publish(Device.availabilityTopic.c_str(), "online");

  String fpsCounter = String(FastLED.getFPS());
  m_mqtt.publish(FPSSensor.statTopic().c_str(), fpsCounter.c_str());
}

void
MQTTTelemetry::callback(char* topic, const char* payload)
{
    setEvent(InputEvent::NetworkActivity);
    if (Lightswitch.isCommandTopic(topic)) {
      deserializeJson(m_json, payload);

      if (m_json.containsKey("state")) {
          if (m_json["state"] == "ON") {
            Log.notice("Turning on power");
            setEvent(InputEvent{InputEvent::SetPower, true});
          } else if (m_json["state"] == "OFF") {
            Log.notice("Turning off power");
            setEvent(InputEvent{InputEvent::SetScene, "Idle"});
            setEvent(InputEvent{InputEvent::SetPower, false});
          }
      }

      if (m_json.containsKey("color")) {
          uint8_t r = m_json["color"]["r"];
          uint8_t g = m_json["color"]["g"];
          uint8_t b = m_json["color"]["b"];
          setEvent(InputEvent{InputEvent::SetColor, CRGB(r, g, b)});
      }

      if (m_json.containsKey("brightness")) {
          setEvent(InputEvent{InputEvent::SetBrightness, m_json["brightness"].as<int>()});
      }

      Log.notice("Event done.");
    } else {
      for(auto scene : Static<Sequencer>::instance()->scenes()) {
        String strName{scene.name};
        MQTTEntity sceneObj {
          "scene",
          Device,
          strName
        };
        if (sceneObj.isCommandTopic(topic)) {
          setEvent(InputEvent{InputEvent::SetScene, scene.name});
          return;
        }
      }
    }
}

void
MQTTTelemetry::s_callback(char* topic, byte* payload, unsigned int length)
{
    strcpy(s_topicBuf, topic);
    memcpy(s_payloadBuf, payload, length);
    s_payloadBuf[std::min(sizeof(s_payloadBuf) - 1, length)] = 0;
    Static<MQTTTelemetry>::instance()->callback(s_topicBuf, s_payloadBuf);
}

char MQTTTelemetry::s_topicBuf[128] = {0};
char MQTTTelemetry::s_payloadBuf[512] = {0};

void
MQTTTelemetry::publishDoc(const char* topic, bool retain)
{
  m_mqtt.beginPublish(topic, measureJson(m_json), retain);
  serializeJson(m_json, m_mqtt);
  m_mqtt.endPublish();
}

void
MQTTTelemetry::publishDoc(const char* topic)
{
  publishDoc(topic, false);
}


STATIC_ALLOC(MQTTTelemetry);
STATIC_TASK(MQTTTelemetry);
